open Ast
open Sast

type mode = Quiet | Verbose (* compiler directives *)

let print_warning mode w = match mode with 
  | Verbose -> print_string("/* WARNING: " ^ w ^ "*/\n")
  | Quiet   -> print_string ""
  
(* Exception types *) 
exception ReturnException of string (* Missing return statement in function *)
exception VariableNotFoundException of string (* Using an undeclared variable *)
exception FunctionNotFoundException of string (* Using an undeclared function *)
exception BinopException of string (* Mismatching types in binop *)
exception OtherException of string (* Misc exception *)
exception AssignmentException of string (* Assigning incompatable things *)
exception FunctionException of string (* Incorrect call to built in functions *)
exception WhileExprException of string (* Incorrect specification of while *)
exception IfExprException of string (* Incorrect specification of if *)
exception ReturnTypeException of string (* Mismatching return statement *)
exception MainNotFoundException of string (* Missing main function *)
exception Error of string (* Placeholder - TODO remove *)

(* Convert data types into string*)
let string_of_type tp = match tp with
      Int -> "Int"
    | String -> "String"
    | Boolean -> "Boolean"
    | Point -> "Point"
    | Curve -> "Curve"
    | Line -> "Line"
    | Ellipse -> "Ellipse"
    | Void -> "Void"

(* Define a symbol table*)
type symbol_table = {
  parent : symbol_table option;
  variables : (string * data_type) list   (* name and type*)
}

(* Define a function table*)
(* all function names are global, no nested funcs allowed, so just build a table of all functions upon program initialization*)
type function_table = {
  functions: (string * data_type * data_type list * stmt) list  (*function name, return type, argument type list, statement *)
}


(* Define the environment *)
type translation_environment = {
  return_type: data_type;     	 (* Function's return type *)
  return_seen: bool; 				 (* If there is a return statement *)
  var_scope: symbol_table;   (* symbol table for vars *)
  fun_scope: function_table; (* list of functions *)
}

(* Add variables' name and type to an envirionment's var_scope*)
let add_var_env env n t = 
  let new_vars = (n,t) :: env.var_scope.variables in
  let new_sym_table = {parent = env.var_scope.parent; variables = new_vars} in
  let new_env = {env with var_scope = new_sym_table} in
  new_env
  
(* Search for variable in the symbol tables *)
let rec find_var_table (var_scope : symbol_table) name =
  try
    List.find (fun (n,_) -> n = name) var_scope.variables
  with Not_found ->
    match var_scope.parent with
        Some(parent) -> find_var_table parent name
      | _ -> raise Not_found
      
(* Search for function in environment*)
let rec find_function (fun_scope : function_table) name =
    List.find (fun (s,_,_,_) -> s = name) fun_scope.functions

(* get type of string literal and int literal *)
let get_int_literal_type l = Int
let get_string_literal_type l = String

(* check whether variable v has type t *)
let check_var_type env v t = 
  let (id, id_t) = find_var_table env.var_scope v in 
  if (id_t = t) then true else false

(* check whether both sides of binop are semantically compatible*)
let check_compatable_types_binop t1 t2 = match (t1,t2) with
    (Int, Int)   			-> true
  | (String, String) 				-> true
  | (Point, Point) 			-> true
  | (Line, Line) 			-> true
  | (Curve, Curve) 					-> true
  | (Ellipse, Ellipse)		-> true
  | _ -> false
  
(* check both sides of assign are semantically compatible*)
let check_compatable_types_assign t1 t2 = match (t1,t2) with
    (Int, Int)   -> true
  | (String, String) 	-> true
  | (Point, Point) -> true
  | (Line, Line) -> true
  | (Ellipse, Ellipse) -> true
  | (Curve, Curve) 	-> true
  | (tp, fp) -> if (tp = fp) then true else false
  
(* Get type of an expression and semantically check it *)
let rec get_type_for_expr env e = match e with
  | IntLiteral(i) -> let t = get_int_literal_type i in t
  | StringLiteral(s) -> let t = get_string_literal_type s in t
  | Id(s) -> let (_, t) = 
      try find_var_table env.var_scope s
      with Not_found -> raise (Error("Undeclared Identifier " ^ s)) in t
  | GetX(e) ->
	let t = get_type_for_expr env e in (if not (t = Point) then raise (Error("This is not a point"))); Int
  | GetY(e) ->
	let t = get_type_for_expr env e in (if not (t = Point) then raise (Error("This is not a point"))); Int  
  | MPoint(e1, e2) -> 
      let t1 = get_type_for_expr env e1 and t2 = get_type_for_expr env e2 in
      (if not (t1 = Int) || not (t2 = Int) 
          then raise (Error("This is not an integer point" ))); Point
  | MLine(e1, e2, e3, e4) -> 
      let t1 = get_type_for_expr env e1 and t2 = get_type_for_expr env e2 and
        t3 = get_type_for_expr env e3 and t4 = get_type_for_expr env e4 in 
      (if not (t1 = Int) || not (t2 = Int) || not (t3 = Int) || not (t4 = Int) 
          then raise (Error("This is not an integer line" ))); Line
  | MEllipse(e1, e2, e3, e4) -> 
      let t1 = get_type_for_expr env e1 and t2 = get_type_for_expr env e2 and
        t3 = get_type_for_expr env e3 and t4 = get_type_for_expr env e4 in 
      (if not (t1 = Int) || not (t2 = Int) || not (t3 = Int) || not (t4 = Int) 
          then raise (Error("This is not an integer ellipse" ))); Ellipse
  | MCurve(e1, e2, e3, e4, e5, e6) -> 
      let t1 = get_type_for_expr env e1 and t2 = get_type_for_expr env e2 and
        t3 = get_type_for_expr env e3 and t4 = get_type_for_expr env e4 and t5 = get_type_for_expr env e5 and t6 = get_type_for_expr env e6 in 
      (if not (t1 = Int) || not (t2 = Int) || not (t3 = Int) || not (t4 = Int) || not (t5 = Int) || not (t6 = Int) 
          then raise (Error("This is not an integer curve" ))); Curve
  | Binop(e1, op, e2) ->
      let t1 = get_type_for_expr env e1 and t2 = get_type_for_expr env e2 in
      (if not (check_compatable_types_binop t1 t2) then
        raise (Error("Mismatch in types for binary operator: " ^ string_of_type t1 ^ ":" ^ string_of_type t2)));
      if op = Equal || op = Neq || op = Less || op = Leq || op = Greater || op = Geq then Boolean else t1
  | Assign(e1, e2) -> 
      let t1 = get_type_for_expr env e1 and t2 = get_type_for_expr env e2 in
      (if not (check_compatable_types_assign t1 t2) then
          raise (Error("Mismatch in types for assignment"))); t1
  | Call(s, el) -> 
      let (fname, fret, farg, fbody) = try find_function env.fun_scope s with Not_found ->
        raise (Error("Undeclared Function " ^ s)) in 
      let el_type = List.map (fun s -> get_type_for_expr env s) el in
      let arg_type = farg in
      (if not (el_type = arg_type) then raise (Error("Mismatching types in function call for: " ^ s))); fret
      
  
(*define semantically checked expression*)
let rec get_sexpr_for_expr env e = let t = get_type_for_expr env e in ExprType(e,t)
  
(*get type of a variable's signature*)
let get_varsig_type vs = let VarSig(t,s) = vs in t

(*get name and type of a variable's signature*)
let get_varsig_name_type vs = let VarSig(t,s) = vs in (s, t)

(*Add function to env and return updated env*)
let add_func_to_env env f =
  let ftable = env.fun_scope in let old_func = ftable.functions in let fun_name = f.fname in 
  let argtype = List.map (fun vs -> get_varsig_type vs) f.formals in let fun_stmt = f.body in
  let fun_ret = f.return in let new_func = (fun_name, fun_ret, argtype, fun_stmt) :: old_func in
  let new_fun_scope = {functions = new_func} in let new_env = {env with fun_scope = new_fun_scope} in
  new_env
  
(* Add a newly initialized var to env and return updated env *)
let add_init_var env ivar = 
  let env = match ivar with
    | NullInit(t, s) -> add_var_env env s t
    | VarInit(t, s, e) -> add_var_env env s t in 
  let (t, env) = match ivar with
    | NullInit(t, s) -> (t, env)
    | VarInit(t, s, e) -> (t, env) in (t, env)
    
(* Combine two environments. it is not used*)
let rec combine_env env lenv =
  let ret_seen = List.fold_left (fun a e -> a && e.return_seen) true lenv in
  let new_env = {env with return_seen = ret_seen} in 
  new_env

(* Get type of a statement and semantically check it *)
let rec get_env_for_stmt env st = match st with 
  | Block(sl) -> 
    let new_var_scope = {parent = Some(env.var_scope); variables = [] } in
    let new_env = {env with var_scope = new_var_scope} in 
    let help(env, acc) s = let (st, e) = get_env_for_stmt env s in (e, st::acc) in
    let (nenv, st) = List.fold_left (fun e s -> help e s) (new_env, []) sl in
    let revst = List.rev st in (SBlock(revst), nenv)
  | Expr(e) -> 
    let _ = get_type_for_expr env e in (SExpr(get_sexpr_for_expr env e), env)
  | Return(e) -> 
    let t = get_type_for_expr env e in 
    (if not (t = env.return_type) then raise (Error("Incompatable Return type")));
    let new_env = {env with return_seen = true} in (SReturn(get_sexpr_for_expr env e), new_env)
  | If(e, s1, s2) -> 
    let t = get_type_for_expr env e in (if not (t = Boolean) then
      	raise (Error("Non-boolean IF predicate")));	
    let (st1, new_env1) = get_env_for_stmt env s1 and (st2, new_env2) = get_env_for_stmt env s2 in
    let ret_seen = (new_env1.return_seen && new_env2.return_seen) in
    let new_env = {env with return_seen = ret_seen} in (SIf((get_sexpr_for_expr env e), st1, st2), new_env)
  | For(e1, e2, e3, s) -> 
    let t1 = get_type_for_expr env e1 and t2 = get_type_for_expr env e2 and t3 = get_type_for_expr env e3 in
    (if not (t1 = Int && t2 = Boolean && t3 = Int) then raise (Error("Bad FOR loop, not integer or boolean bounds")));
    (if not (t1 = t3) then raise (Error("Bad FOR loop, can not assign to iterator")));
    let (st, new_env) = get_env_for_stmt env s in 
    (SFor((get_sexpr_for_expr env e1), (get_sexpr_for_expr env e2), (get_sexpr_for_expr env e3), st), new_env)
  | While(e, s) -> 
    let t = get_type_for_expr env e in
    (if not (t = Boolean) then raise (Error("Bad WHILE loop, not boolean bounds")));
    let (st, new_env) = get_env_for_stmt env s in (SWhile((get_sexpr_for_expr env e), st), new_env)
  | Print(e) -> 
    let _ = get_type_for_expr env e in (SPrint(get_sexpr_for_expr env e), env)
  | NoElseMark -> (SNoElseMark, env)
  | VarDecl(i) -> 
    let (si, new_env) = add_init_var env i in (SVarDecl(SInit(i)), new_env)
  | Draw(e) -> let _ = get_type_for_expr env e in (SDraw(get_sexpr_for_expr env e), env)
    
(* Add all global vars and functions to our environment *)
let initialize_globals env constr = match constr with
  | GlobalVar(i) ->
      let (_, new_env) = add_init_var env i in
      new_env
  | FuncDef(f) ->
      let new_env = add_func_to_env env f in
      new_env
      
(* Check that return statements are in order, etc. *)
let check_final_env env = 
  (if (false = env.return_seen && env.return_type <> Void) then
    raise (Error("Missing Return Statement")));
  true  
  
(* Semantic checking on a function *)
let check_funcs env ct = match ct with
  | GlobalVar(v) ->
      let (t, _) = add_init_var env v in
      SGlobalVar(SInit(v))
  | FuncDef(f) ->
      let new_variables = List.fold_left (fun a vs -> (get_varsig_name_type vs)::a) [] f.formals in
      let new_var_scope = {parent = Some(env.var_scope); variables = new_variables} in
      let new_env = {return_type = f.return; return_seen = false; var_scope = new_var_scope; fun_scope = env.fun_scope} in
      let (stbody, final_env) = get_env_for_stmt new_env f.body in
      let _ = check_final_env final_env in
      let sfunc = {sreturn = f.return; sfname = f.fname; sformals = f.formals; sbody = stbody} in
      SFuncDef(sfunc)
      
(* Empty initializations for tables/environments *)
let empty_symbol_table = {parent = None; variables = []}
let empty_function_table = {functions = []}
let empty_env = {return_type = Void; return_seen = false; var_scope = empty_symbol_table; fun_scope = empty_function_table}


(* MAIN RUNNER FUNCTION *)
let translate prog =
  let env = List.fold_left (fun env ct -> initialize_globals env ct) empty_env prog in
  let res = List.map (fun ct -> check_funcs env ct) prog in
  res



    

    

    


  


  
  
          
  
    
  




















